/* This is a simplified interface to the dynamic-linking routines.
 * It must deal with the fact that the dynamic loader likes to load
 * code at high addresses ( > 1sagan), but we can only deal with low
 * addresses (< 1 sagan).  We solve this by building a thunk in low
 * memory that calls the actual routine in high memory, and we return
 * the address of that thunk.  Ugly, I know...
 */

#include <stdio.h>
#include <errno.h>
#include <dlfcn.h>

extern FILE * x86_logfile;

extern int	trace_callbacks;
struct tdebug {
	unsigned int (*routine)();
	char name[64];
};

#define GLOBAL_SCOPE	1

unsigned int axp_call_thunk(unsigned int a0, unsigned int a1, unsigned int a2,
			    unsigned int a3, unsigned int a4,
			    struct tdebug * td)
{
    unsigned int retval;
    if(strcmp(td->name, "getenv") == 0) {
        fprintf(x86_logfile, "Calling thunk: %s(%s, 0x%x, 0x%x, 0x%x, 0x%x)\n",
		td->name, a0, a1, a2, a3, a4);
    }
    else {
        fprintf(x86_logfile, "Calling thunk: %s(0x%x, 0x%x, 0x%x, 0x%x, 0x%x)\n",
		td->name, a0, a1, a2, a3, a4);
    }
    fflush(x86_logfile);
    retval = (td->routine)(a0, a1, a2, a3, a4);
    if(strcmp(td->name, "getenv") == 0) {
        fprintf(x86_logfile, "  ...%s returns %s\n", td->name, retval);
    }
    else {
        fprintf(x86_logfile, "  ...%s returns 0x%x\n", td->name, retval);
    }
    fflush(x86_logfile);
    return(retval);
}


void * axp_dlopen(char * filename)
{
    void *	handle;
    void **	handlep;

    /* For routines that are likely already part of the em86
     * executable, just use the global scope...
     */
    if( (strncmp(filename, "libc.so", 7) == 0) ||
        (strncmp(filename, "libm.so", 7) == 0) ) {
	handle = GLOBAL_SCOPE;
    }
    else {
	handle = dlopen(filename, RTLD_NOW);
	if(!handle) {
	    fprintf(x86_logfile, "EM86: axp_dlopen of %s failed: %s\n", 
			    filename, dlerror());
	    _exit(-alpha_to_intel_errno(ENOENT));
	}
    }

    /* Allocate a pointer in low memory to refer to this handle */
    handlep = (void *)malloc(sizeof(void *));
    *handlep = handle;
#ifdef VERBOSE
    fprintf(x86_logfile, "axp_dlopen(%s): handle = 0x%lx, handlep = 0x%lx\n",
		filename, handle, handlep);
#endif
    return(handlep);
}

int axp_dlclose(void ** handlep)
{
    if(*handlep == GLOBAL_SCOPE) {
	return(0);
    }
    else {
        return(dlclose(*handlep));
    }
}

void * axp_dlsym(void **handlep, char *symbol)
{
    void *symval;
    struct tdebug *	td;
    unsigned int *	thunk_thunk;

    if(*handlep == GLOBAL_SCOPE) {
	symval = dlsym((void *)0, symbol);
    }
    else {
	symval = dlsym(*handlep, symbol);
    }

    if(!symval) {
	fprintf(x86_logfile, "EM86: axp_dlsym failed to locate %s: %s\n",
		symbol, dlerror());
	_exit(-1);
    }
#ifdef VERBOSE
    fprintf(x86_logfile, "axp_dlsym(%s): routine @0x%lx\n", symbol, symval);
#endif

    if(trace_callbacks) {
	/* Build a thunk to call through the printout routine... */
	td = (struct tdebug *)malloc(sizeof(struct tdebug));
	thunk_thunk = (unsigned int *)malloc(8 * sizeof(unsigned int));

	td->routine = symval;
	strcpy(td->name, symbol);
	thunk_thunk[0] = 0xc2a00001;	/* br $a5,.+4 */
	thunk_thunk[1] = td;
	thunk_thunk[2] = 0xa2b50000;	/* ldl $a5,0($a5) */
	thunk_thunk[3] = 0xc3600001;	/* br $pv,.+4 */
	thunk_thunk[4] = axp_call_thunk;
	thunk_thunk[5] = 0xa37b0000;	/* ldl $pv,0($pv) */
	thunk_thunk[6] = 0x6bfb4000;	/* jsr $31,($pv) */
	thunk_thunk[7] = 0;

	return(thunk_thunk);
    }
    else {
	return(symval);
    }
}

void * axp_build_thunk(void** handlep, char *symbol, unsigned int * code)
{
    void *	symval;
    unsigned int *	thunk;
    int		thunk_size;
    int		i;
    int		user_code_size;

    symval = dlsym(*handlep, symbol);

    /* The thunk looks like this:
     *
     *	thunk:	<optional-user-code>
     *	        br  r27,1f
     *		[quadword-address-of-real-routine]
     *	    1:  jmp zero,(r27)
     *		nop
     */
    /* Calculate the size of the user-specified code (if any) */
    if(code) {
	for(i = 0; code[i]; i++) ;
	user_code_size = i;
    }
    else {
	user_code_size = 0;
    }
	    
    /* How much memory to allocate for the thunk?
     *    user_code_size * 4 PLUS
     *    4 for the br instruction, PLUS
     *    8 for the routine address, PLUS
     *    4 for the jmp instruction, PLUS
     *    8 for nops that might be needed for alignment.
     *    4 for a nop after the thunk call (needed by some
     *      signalling mechanisms)
     *    4 for a halt instruction after the thunk
     *
     * i.e. (user_code_size * 4) + 32
     */

    thunk_size = (user_code_size * 4) + 32;

    /* Round thunk_size up to a quadword boundary */
    thunk_size = (thunk_size + 7) & ~7;
  
    thunk = (unsigned int *)malloc(thunk_size);
    for(i = 0; i < user_code_size; i++) {
	thunk[i] = code[i];
    }
    
    /* if i is even, we must add a nop so that the quadword address
     * is also even.
     */
    if((i & 1) == 0) {
        thunk[i++] = 0x47ff041f;		/* nop */
    }
    thunk[i++] = 0xc3600002;		/* br r27,.+8 */
    *((unsigned long *)(&thunk[i])) = (unsigned long)symval;
    i += 2;
    thunk[i++] = 0xa77b0000;		/* ldq r27,0(r27) */
    thunk[i++] = 0x6bfb0000;		/* jmp zero,(r27) */
    thunk[i++] = 0x47ff041f;		/* nop */
    thunk[i++] = 0;			/* halt */

    if((i * 4) > thunk_size) {
	fprintf(x86_logfile, "PANIC! thunk_size = %d, used %d bytes\n",
		thunk_size, i*4);
    }
#ifdef VERBOSE
    fprintf(x86_logfile, "axp_dlsym(%s): routine @0x%lx, thunk @0x%lx\n",
		    symbol, symval, thunk);
#endif
    return(thunk);
}

