#include <sys/types.h>
#include <sys/mman.h>
#include <signal.h>
#include "elf.h"
#include "linux_bin.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

#include "x86_cpu.h"

/* Cribbed from personality.h */
#define PER_LINUX_EM86          0x0800000

#define ALPHA_PAGE_SIZE	0x2000
#define ALPHA_PAGE_MASK 0xffffe000
/* This was 0xc0000000, but we ran into sign-extension problems in the
 * FX emulator....
 */
#define INITIAL_SP	0x80000000
#define INITIAL_STACK_SIZE	0x40000

FILE * x86_logfile;
X86_CPU *	current_cpu;

int	debug_mmap = 0;
int	die_on_segv = 0;
int	trace_callbacks = 0;
int	trace_instructions = 0;

unsigned long x86_stack_size;

extern "C" {
    int elf_exec(char *, char **, char **, struct pt_regs *, struct image_info *);
    void x86_set_brk(char *);
    void x86_syscall_debug(int);
    void x86_signal_handler_init();
    void x86_process_pending_signals();
    void x86_full_disassemble(unsigned int);
}

extern int x86_sigmask;

void sig_handler(int);

unsigned long	stktop;

main(int argc, char ** argv, char ** envp)
{
    X86_CPU	x86_cpu;
    struct pt_regs regs;
    struct image_info info;
    char *	syscall_trace;

    personality(PER_LINUX_EM86);

    if(argc < 2) {
	printf("Usage: %s executable-file [args ...]\n", argv[0]);
	exit(0);
    }

    if(strcmp(argv[1], "--help") == 0) {
	char *cp;
	printf("Usage: %s executable-file [args ...]\n", argv[0]);
	printf("The following environment variables can be used to control execution:\n");
	printf("    X86_TRACE_SYSCALLS - Traces system calls\n");
	printf("        e.g. X86_TRACE_SYSCALLS=open,read,write\n");
	printf("             X86_TRACE_SYSCALLS=all,-select\n");
	printf("    X86_TRACE_SIGNALS - Traces signal delivery\n");
	printf("        e.g. X86_TRACE_SIGNALS=SIGSEGV,SIGINT\n");
	printf("             X86_TRACE_SIGNALS=all,-SIGALRM\n");
	printf("    X86_STACK_SIZE - Specify initial stack size in megabytes\n");
	printf("    X86_SEGV_FATAL - Die if any SIGSEGV received\n");
	printf("    X86_DEBUG_MMAP - Traces memory-space management\n");
	printf("    X86_LOGFILE - File to receive all output from em86\n");
	printf("    X86_TRACE_CALLBACKS - Trace calls between x86 and native environments\n");
	printf("    X86_TRACE_INSTRUCTIONS - Trace every instruction (SLOW!)\n");
	printf("    X86_SAVE_TRACE - Save last 64 instructions for post-mortem trace\n");
#ifdef USE_FX_CPU
	printf("    FX_INDEX_FILE - Location of the 'linux486.idx' file\n");
#endif
#ifdef USE_BOCHS_CPU
	printf("    BOCHS_DEBUG_CPU - Trace all x86 instructions\n"); 
#endif
	printf("Currently defined environment variables:\n");
	if(cp = getenv("X86_TRACE_SYSCALLS")) {
	    printf("  X86_TRACE_SYSCALLS = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_TRACE_SIGNALS")) {
	    printf("  X86_TRACE_SIGNALS = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_STACK_SIZE")) {
	    printf("  X86_STACK_SIZE = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_SEGV_FATAL")) {
	    printf("  X86_SEGV_FATAL = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_DEBUG_MMAP")) {
	    printf("  X86_DEBUG_MMAP = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_LOGFILE")) {
	    printf("  X86_LOGFILE = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_TRACE_CALLBACKS")) {
	    printf("  X86_TRACE_CALLBACKS = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_TRACE_INSTRUCTIONS")) {
	    printf("  X86_TRACE_INSTRUCTIONS = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("X86_SAVE_TRACE")) {
	    printf("  X86_SAVE_TRACE = %s (0x%lx)\n", cp, cp);
	}
	if(cp = getenv("FX_INDEX_FILE")) {
	    printf("  FX_INDEX_FILE = %s (0x%lx)\n", cp, cp);
	}
	printf("argc = 0x%x, argv = 0x%lx, envp = 0x%lx\n", argc, argv, envp);

	exit(0);
    }

    /* Check and see if we're running on a properly-patched kernel... */
    if( ((unsigned long)(&regs)) >> 31 ) {
	printf("Your kernel does not support 31-bit addressing!\n");
	printf("Please apply the EM86 kernel patch and try again.\n");
	exit(0);
    }

    /* Specify preload library */
    if(!getenv("X86_NO_PRELOAD")) {
      if (!getenv("X86_LD_PRELOAD")) {
	setenv("X86_LD_PRELOAD", "/usr/x86/lib/libc-thunk.so", 1);
      }
      else {
	char *	tmpstr = malloc(strlen(getenv("X86_LD_PRELOAD"))+32);
	strcpy(tmpstr, getenv("X86_LD_PRELOAD"));
	setenv ("X86_LD_PRELOAD", strcat(tmpstr,
					 ":/usr/x86/lib/libc-thunk.so"), 1);
      }
    }

    /* Zero out regs */
    memset(&regs, 0, sizeof(struct pt_regs));

    /* Zero out image_info */
    memset(&info, 0, sizeof(struct image_info));

    if(getenv("X86_DEBUG_MMAP")) {
	debug_mmap = 1;
    }

    if(getenv("X86_TRACE_CALLBACKS")) {
	trace_callbacks = 1;
    }
    if(getenv("X86_SEGV_FATAL")) {
	die_on_segv = 1;
    }

    if(getenv("X86_STACK_SIZE")) {
        x86_stack_size = atoi(getenv("X86_STACK_SIZE")) * 1024;
    }

    if(getenv("X86_LOGFILE")) {
	x86_logfile = fopen(getenv("X86_LOGFILE"), "a");
	if(x86_logfile == NULL) {
	    fprintf(stderr, "Cannot open %s as logfile\n", 
			getenv("X86_LOGFILE"));
	    x86_logfile = stderr;
	}
	x86_cpu.set_logfile(x86_logfile);
    }
    else {
	x86_logfile = stderr;
	x86_cpu.set_logfile(x86_logfile);
    }

    /* Enable syscall debugging if desired... */
    syscall_trace = getenv("X86_TRACE_SYSCALLS");
    if(syscall_trace) {
	char	*syscall_name;

	syscall_name = strtok(syscall_trace, ",");
	while(syscall_name) {
	    if(strcmp(syscall_name, "all") == 0) {
		x86_syscall_trace_all();
	    }
	    else {
		if(*syscall_name == '-') {
		    x86_syscall_untrace(syscall_name+1);
		}
		else {
		    x86_syscall_trace(syscall_name);
		}
	    }
	    syscall_name = strtok(NULL, ",");
	}
    }


    if(elf_exec(argv[1], argv+1, __environ, &regs, &info) != 0) {
	printf("Error loading %s\n", argv[1]);
	exit(-1);
    }

    x86_set_brk(info.brk);

    x86_signal_handler_init();


    /* Okay; program is loaded... now try to run it! */
    x86_cpu.init_cpu( regs.eax, regs.ebx, regs.ecx, regs.edx,
    		      regs.esi, regs.edi, regs.ebp, regs.esp,
		      regs.eip);
    current_cpu = &x86_cpu;

    if(getenv("X86_TRACE_INSTRUCTIONS")) {
	trace_instructions = 1; 
    } 

    if(getenv("X86_SAVE_TRACE")) {
	current_cpu->start_trace();
    }

    for(;;) {
	
	if(trace_instructions) {
	    fprintf(x86_logfile, "Instruction trace: ");
	    fprintf(x86_logfile, 
			"EAX: %08x  EBX: %08x  ECX: %08x  EDX: %08x\n",
			current_cpu->get_EAX(), current_cpu->get_EBX(),
			current_cpu->get_ECX(), current_cpu->get_EDX());
	    fprintf(x86_logfile, 
			"                   ESI: %08x  EDI: %08x  EBP: %08x  ESP: %08x\n",
			current_cpu->get_ESI(), current_cpu->get_EDI(),
			current_cpu->get_EBP(), current_cpu->get_ESP());
	    x86_full_disassemble(current_cpu->get_EIP());
	    fprintf(x86_logfile, "\n");
	    current_cpu->instruction_break(2);
	}

        x86_cpu.cpu_loop();

	/* If we return from the loop, it's quite likely because
	 * we caught a signal...
	 */
	if(x86_sigmask) {
	    x86_process_pending_signals();
	}
    }
}

unsigned long instruction_count = 0;

extern "C" {
    void x86_process_pending_signals();
}

/* This routine is called by the Bochs emulator after every
 * instruction.  In addition to keeping statistics, we'll use
 * this as a useful hook to do deferred signal processing....
 */
#ifdef USE_BOCHS_CPU
void bx_instruction_tick(unsigned int eip)
{
    instruction_count++;
    x86_process_pending_signals();
}
#endif

extern "C" void em86_exit(void)
{
#ifdef USE_BOCHS_CPU
    printf("%d instructions emulated\n", instruction_count);
#endif
}

/* This routine is used in certain thunking paths when Alpha code
 * need to invoke an Intel function...
 */
extern "C" unsigned int em86_call_emulated_routine(unsigned int a0,
		unsigned int a1, unsigned int a2, unsigned int a3,
		unsigned int a4, void * addr)
{
    unsigned int	esp, eip;
    unsigned int	*stkp;

    if(!current_cpu->registers_homed) {
	fprintf(x86_logfile, 
		"AIEEE! em86_call_emulated_routine, regs not homed!\n");
    }
    /* Find out the current emulated stack pointer */
    esp = current_cpu->get_ESP();

    /* Longword-align the stack, then push on two longwords... */
    stkp = (unsigned int *)(esp & ~3);
    stkp -= 7;

    /* The first longword points to the last.  The last longword is
     * the EM86 "return-from-emulator" instruction.  This way, when the
     * called function issues a "ret" instruction, the return PC points
     * to the "return-from-emulator" instruction on the stack, which
     * then causes cpu_loop() to return.
     *
     * The intervening longwords are up to five integer arguments to 
     * the called function...
     */
    stkp[0] = (unsigned int)( ((unsigned long)(&(stkp[6]))) & 0xffffffff);
    stkp[0] += 3;
    stkp[6] = 0xd6000000;

    stkp[1] = a0;
    stkp[2] = a1;
    stkp[3] = a2;
    stkp[4] = a3;
    stkp[5] = a4;

    /* All set... call the emulator. */
    current_cpu->set_ESP((unsigned int)stkp);
    current_cpu->set_EIP((unsigned int)addr);
    if(trace_callbacks) {
	fprintf(x86_logfile, "Calling emulated routine @0x%lx(0x%x, 0x%x, 0x%x, 0x%x, 0x%x)\n",
		addr, a0, a1, a2, a3, a4);
    }
    current_cpu->cpu_loop();

    /* Put the stack back where we found it... */
    current_cpu->set_ESP(esp);

    if(trace_callbacks) {
	fprintf(x86_logfile, "...returns 0x%x\n", current_cpu->get_EAX());
    }

    /* Return value from an Intel function is in EAX */
    return(current_cpu->get_EAX());
}

extern "C" void
EM486_lock_memory() {
}

extern "C" void
EM486_unlock_memory() {
}

extern "C" void
em86_start_instruction_tracing()
{
    current_cpu->instruction_break(1);
    trace_instructions=1;
}

